@@ -1,12 +1,15 @@
+# frozen_string_literal: true
 # date.rb: Written by Tadayoshi Funaba 1998-2011
 
 require 'date_core'
 
 class Date
 
-  class Infinity < Numeric # :nodoc:
+  def infinite?
+    false
+  end
 
-    include Comparable
+  class Infinity < Numeric # :nodoc:
 
     def initialize(d=1) @d = d <=> 0 end
 
@@ -14,26 +17,28 @@ class Date
 
     protected :d
 
-    def zero? () false end
-    def finite? () false end
-    def infinite? () d.nonzero? end
-    def nan? () d.zero? end
+    def zero?() false end
+    def finite?() false end
+    def infinite?() d.nonzero? end
+    def nan?() d.zero? end
 
     def abs() self.class.new end
 
-    def -@ () self.class.new(-d) end
-    def +@ () self.class.new(+d) end
+    def -@() self.class.new(-d) end
+    def +@() self.class.new(+d) end
 
-    def <=> (other)
+    def <=>(other)
       case other
       when Infinity; return d <=> other.d
+      when Float::INFINITY; return d <=> 1
+      when -Float::INFINITY; return d <=> -1
       when Numeric; return d
       else
-	begin
-	  l, r = other.coerce(self)
-	  return l <=> r
-	rescue NoMethodError
-	end
+        begin
+          l, r = other.coerce(self)
+          return l <=> r
+        rescue NoMethodError
+        end
       end
       nil
     end
@@ -42,16 +47,16 @@ class Date
       case other
       when Numeric; return -d, d
       else
-	super
+        super
       end
     end
 
     def to_f
       return 0 if @d == 0
       if @d > 0
-	Float::INFINITY
+        Float::INFINITY
       else
-	-Float::INFINITY
+        -Float::INFINITY
       end
     end
 
